from enum import StrEnum
from typing import Optional, Annotated
from typing_extensions import Doc

from settings_models._combat import SettingsModel, Field
from settings_models.validators import hhmm_validator


class GateMode(StrEnum):
    """
    How the gate should behave
    """
    automatic = "automatic"
    permanent_open = "permanent_open"
    permanent_close = "permanent_close"
    standard = "standard"


class DayMode(SettingsModel):
    """
    If day mode (everyone can enter) is enabled and at what time of day
    """
    enabled: bool
    start: str = Field(default="00:00", description="Start of day mode in format HH:MM local time")
    end: str = Field(default="00:00", description="End of day mode in format HH:MM local time")

    start_validator = hhmm_validator("start")
    end_validator = hhmm_validator("end")


GateModeType = Annotated[Optional[GateMode], Doc(GateMode.__doc__)]
DayModeType = Annotated[Optional[DayMode], Doc(DayMode.__doc__)]



from typing import TypeVar, Type
from pydantic import TypeAdapter

T = TypeVar("T")


class Setting:
    settings_model: TypeAdapter[T]
    settings_doc: str
    scoped: bool

    def __init__(self, settings_model: Type[T], settings_doc: str = "", scoped: bool = False):
        self.settings_doc = settings_doc or settings_model.__doc__
        self.settings_model = TypeAdapter(settings_model)
        self.scoped = scoped



from typing import Optional

from pydantic import model_validator

from settings_models._combat import SettingsModel, Field


class IntercomSettings(SettingsModel):
    """
    Settings for intercom on kiosk devices
    """
    enabled: bool = Field(..., description="If intercom enabled on kiosk devices")
    phone_number: Optional[str] = Field(default=None, description="Phone number for intercom emergency calls")

    @model_validator(mode="after")
    def data_validation(self):
        if self.enabled and self.phone_number is None:
            raise ValueError("phone_number must be set if intercom is enabled")
        return self



from datetime import datetime
from typing import Optional, TypeAlias, Annotated
from typing_extensions import Doc

from settings_models._combat import SettingsModel, Field
from settings_models.validators import timezone_validator, deprecated_validator


class EnforcementSettings(SettingsModel):
    """
    Settings for enforcement service
    """
    payment_deadline_hours: int = Field(..., gt=0, le=960,
                                        description="How long someone can pay after exiting parking lot "
                                                    "(not including generally applied additional day)")
    strictness: int = Field(default=2, deprecated="The field `EnforcementSettings.strictness` is deprecated "
                                                  "and will be removed in a future version.",
                            description="Strictness level for enforcement. Values to be defined.")
    enabled: bool = Field(..., description="If enforcement is enabled")
    ai_enabled: bool = Field(default=True, description="If the Arivo AI is enabled")
    last_edited: datetime = Field(..., description="Last time (UTC) when enabled setting was changed")

    last_edited_validator = timezone_validator("last_edited")
    strictness_validator = deprecated_validator("strictness")


EnforcementSettingsType = Annotated[Optional[EnforcementSettings], Doc(EnforcementSettings.__doc__)]



import re
from decimal import Decimal
from typing import Union, Type, Optional, Any, TypeVar

import orjson
from dateutil.tz import gettz
from pydantic import TypeAdapter

from settings_models.doc import Setting
from settings_models.settings.assets import AssetVisualDefinitions
from settings_models.settings.common import SettingsModel, Gates, Rates, ParkingAreas, Parksettings, CostGroups, \
    PrivacySettings, BillingSettings, Location, SupportSettings, RefsModel, GarageSettings, GarageName, Urls
from settings_models.settings.device_keys import PairingKey
from settings_models.settings.enforcement import EnforcementSettingsType
from settings_models.settings.gate_control import GateModeType, DayModeType
from settings_models.settings.intercom import IntercomSettings
from settings_models.settings.io import SignalDefinitions, InputDefinitions

T = TypeVar('T')
RefsModelAdapter = TypeAdapter(RefsModel)

_model_mapping = {
    "common/gates": Setting(Gates, "Has one entry for each gate of the garage."),
    "common/rates": Setting(Rates, "Defines the parking costs."),
    "common/parking_areas2": Setting(ParkingAreas, "Defines the physical areas of parking lot by specifying the gates."
                                                   "Also specifies settings for shortterm users."),
    "common/parksettings2": Setting(Parksettings, "Defines the terms and conditions for parking sessions of "
                                                  "registered users (not including registered shortterm users)."),
    "common/privacy_settings": Setting(PrivacySettings),
    "common/cost_groups": Setting(CostGroups, "Defines the costs for different activities."),
    "common/currency": Setting(str, "Defines the currency in which the parking costs are calculated."),
    "common/language": Setting(str, "Defines the language used for the user interface."),
    "common/timezone": Setting(str, "Defines the timezone in which the garage is located."),
    "common/garage_settings": Setting(GarageSettings),
    "common/location": Setting(Location),
    "common/billing": Setting(BillingSettings),
    "common/support": Setting(SupportSettings),
    "common/garage_name": Setting(GarageName),
    "common/urls": Setting(Urls),
    "gate_control/mode": Setting(GateModeType, scoped=True),
    "gate_control/day_mode": Setting(DayModeType),
    "enforcement/basic_settings": Setting(EnforcementSettingsType),
    "intercom/basic_settings": Setting(IntercomSettings),
    "device_keys/pairing": Setting(PairingKey),
    "feature_flags": Setting(dict, "A dictionary of custom values that can be used to enable or "
                                   "disable features in various places."),
    "io/signals": Setting(SignalDefinitions, "Defines the available signals."),
    "io/inputs": Setting(InputDefinitions, "Defines the available inputs."),
    "assets/visuals": Setting(AssetVisualDefinitions, "Defines the available visual assets."),

}


def _get_type(key: str, custom_type: Optional[Type[T]]) -> TypeAdapter[T]:
    if custom_type:
        return TypeAdapter(custom_type)
    if "/refs/" in key:
        return RefsModelAdapter
    if key not in _model_mapping:
        raise KeyError(f"No model for setting {key} found")
    return _model_mapping[key].settings_model


def _key_equal_id_in_value(k: str, v: Any, id_field: str = "id"):
    if k != getattr(v, id_field):
        raise ValueError(f"key must match field {id_field}")


def _special_validation(key: str, obj: Any):
    if key in ["common/rates", "common/parking_areas2", "common/parksettings2", "common/cost_groups"]:
        for k, v in obj.items():  # those settings are always dicts
            _key_equal_id_in_value(k, v)
    elif key == "common/gates":
        for k, v in obj.items():  # those settings are always dicts
            _key_equal_id_in_value(k, v, "gate")
            if v.opposite_gate is not None:
                if v.opposite_gate not in obj or obj[v.opposite_gate].opposite_gate != k:
                    raise ValueError(f"gate {k} and {v.opposite_gate} must reference each other")
    elif key == "common/timezone":
        if gettz(obj) is None:
            raise ValueError(f"timezone {obj} unknown")
    elif key == "common/language":
        if re.search(r"^[a-z]{2}(-[A-Z]{2})?$", obj) is None:
            raise ValueError(f"{obj} is no valid language code")
    elif key == "common/currency":
        if len(obj) != 3 or not obj.isalpha() or not obj.isupper():
            raise ValueError("Currency must be 3 uppercase letters")
    elif key == "io/signals":
        for k, v in obj.items():
            _key_equal_id_in_value(k, v, "signal")
    elif key == "io/inputs":
        for k, v in obj.items():
            _key_equal_id_in_value(k, v, "input")
    elif key == "assets/visuals":
        for k, v in obj.items():
            _key_equal_id_in_value(k, v, "type")


def parse_setting_from_obj(key: str, value: Any, custom_type: Optional[Type[T]] = None) -> Any:
    """
    Parses setting into model of given key performing model validations and some additional validations for basic
    datatypes.
    :param key: Key of setting: is used to select the SettingsModel
    :param value: Settings value as json decoded object (or even already parsed object)
    :param custom_type: If your service uses only some field within a setting, a less specific model can be passed
    :return: Value parsed as correct model for given key
    :raises: ValueError if validation of setting fails
    """
    target_type = _get_type(key, custom_type)
    obj = target_type.validate_python(value)
    if not custom_type:
        _special_validation(key, obj)
    return obj


def parse_setting(key: str, value: Union[str, bytes], custom_type: Optional[Type[T]] = None) -> Any:
    """
    Parses setting into model of given key performing model validations and some additional validations for basic
    datatypes.
    :param key: Key of setting: is used to select the SettingsModel
    :param value: Settings value as string or byte array. Must be JSON decodable!
    :param custom_type: If your service uses only some field within a setting, a less specific model can be passed
    :return: Value parsed as correct model for given key
    :raises: ValueError if validation of setting fails
    """
    try:
        value_obj = orjson.loads(value)
    except orjson.JSONDecodeError as e:
        raise ValueError(f"{key} requires json decodable value") from e
    return parse_setting_from_obj(key, value_obj, custom_type)


def _default(obj):
    if isinstance(obj, SettingsModel):
        return obj.model_dump()
    if isinstance(obj, Decimal):
        return str(obj)
    raise TypeError


_ESCAPE_ASCII = re.compile(r'([^ -~])')


def _replace(match):
    s = match.group(0)
    n = ord(s)
    if n < 0x10000:
        return '\\u%04x' % (n,)
    else:
        # surrogate pair
        n -= 0x10000
        s1 = 0xd800 | ((n >> 10) & 0x3ff)
        s2 = 0xdc00 | (n & 0x3ff)
        return '\\u%04x\\u%04x' % (s1, s2)


def _unicode_escape(data):
    return _ESCAPE_ASCII.sub(_replace, data)


def dump_setting(obj: Any) -> bytes:
    """
    Serializes settings model into bytearray for sending messages and writing into databases
    :param obj: setting as parsed model
    :return: byte array of serialized setting
    """
    return _unicode_escape(orjson.dumps(obj, default=_default).decode()).encode("ascii")



from datetime import datetime
from enum import StrEnum
from typing import Dict, List, Optional, Union

from pydantic import constr, model_validator, AfterValidator
from typing_extensions import Annotated

from settings_models._combat import SettingsModel, Field
from settings_models.validators import uuid_validator, domain_validator, account_id_range_validator


class GateType(StrEnum):
    """
    Types of a gate. This defines the usage and the general behavior of a gate
    """

    entry = "entry"
    """Entry gate: Used where vehicles typically or exclusively enter the parking lot"""

    exit = "exit"
    """Exit gate: Used where vehicles typically or exclusively exit the parking lot"""

    bidirectional = "bidirectional"
    """Bidirectional gate: Used where vehicles typically enter AND exit the parking lot"""

    door = "door"
    """
    Door: Gate without LPR where people can walk into the parking lot.
    No check-in or check-out, only access check
    """

    transit_access_check = "transit_access_check"
    """
    Gate with barrier where vehicles can pass depending on access rights specified by the gate list in parksetting or
    shortterm_gates in parking_area. This is a well defined functionality of the controller.
    """

    transit_custom = "transit_custom"
    """
    Gates with custom behavior. These gates might or might not have a barrier. Examples are:
      - Passage check gates (permanent parkers must pass this gate or gets fined)
      - Gate with shortterm count check (e.g. only 10 shortterm parkers are allowed at the same time)
      - automatic gates (e.g. exit of restricted subarea)
    Behavior of those gates is not implemented by the controller
    """

    coupled_lane = "coupled_lane"
    """
    Gates which open always together with another lane (e.g. garage door after entry barrier) but can also be controlled
    independently, e.g. perma open
    """

    check_in_out = "check_in_out"
    """
    Freeflow entries, exits or bidirectional lanes without hardware. We only checkin or checkout license plates here.
    """

    transit_passthrough = "transit_passthrough"
    """
    Transit lanes which are only used to detect passing license plates,
    e.g. to change parking rate or to check if someone parked in the correct area
    """


class RefsModel(SettingsModel):
    """
    Model for used keys of a Dict setting like parksettings
    """
    name: str = Field(..., description="Human readable name of the service using the keys for error messages",
                      max_length=126)
    keys: List[str] = Field(..., description="Keys referenced by the service")


class Gate(SettingsModel):
    """
    Definition of a Gate
    """
    gate: constr(min_length=1, max_length=126, pattern=r"^[a-z0-9\-_]+$") = \
        Field(..., description="Technical name of gate")
    name: str = Field(..., description="Name for UIs", max_length=126)
    type: GateType = Field(..., description="Type of the gate. Defines the usage and general behavior")
    opposite_gate: Optional[str] = Field(default=None, description="Name of the opposite gate for bidirectional gates")
    is_wide: Optional[bool] = Field(default=None,
                                    description="This field in the gate settings model indicates whether the gate is "
                                                "wide. For gates of type entry and exit in Freeflow mode, the value "
                                                "must be either true or false. For all other gate types, the value must "
                                                "be null.")


class Rate(SettingsModel):
    """
    Definition of a parking rate
    """
    rate_yaml: str = Field(..., description="Parking rate config: https://gitlab.com/accessio/onegate/parking-rates")
    name: str = Field(..., description="Name for UIs", max_length=126)
    id: str = Field(..., description="id for referencing of this rate")

    id_validator = uuid_validator("id")


class CommandCostEntryTypes(StrEnum):
    """
    Types of command cost entries: Those are used to cost entries in a parking session, e.g. via parksettings
    """

    static_cost = "static_cost"
    """static cost immediately booked"""

    pending_static_cost = "pending_static_cost"
    """static cost booked at exit"""

    rate_change = "rate_change"
    """for rate changes that don't generate revenue reduction entry"""

    rate_validation = "rate_validation"
    """for rate changes that generate revenue reduction entry (usually reducing costs)"""

    partial_rate_change = "partial_rate_change"
    """for rate changes which freeze costs until now"""

    amount_validation = "amount_validation"
    """validation for up to a certain amount"""

    time_validation = "time_validation"
    """validation for a certain time"""

    payment = "payment"
    """payment"""

    cost_validation = "cost_validation"
    """generates a validation (gen_revenue_reduction) for open costs in group"""

    disable_costs = "disable_costs"
    """cancels costs in group (either disable before booking or gen_cancellation)"""

    cancel_remaining_costs = "cancel_remaining_costs"
    """generates cancellation over remaining amount for each group"""

    payoff_costs = "payoff_costs"
    """generates pay off over (set to 0 remaining when session not editable) for each group"""

    free_timerange = "free_timerange"
    """
    instead of calculating dynamic costs from start to end, 
    we calculate them from start to free_timerange start + free_timerange end to end
    """


class CostEntryData(SettingsModel):
    """
    Defines a cost entry command, which is for example set at entering the parking lot
    """
    entry_type: CommandCostEntryTypes = Field(..., description="Cost entry command type")
    value: Optional[Union[int, str, Dict[str, Optional[datetime]]]] = (
        Field(default=None, description="value: Dependent on entry_type it's an int, a str or None"))
    group: Optional[str] = Field(default=None, description="Cost group the command cost entry is applied on. None for "
                                                           "global")
    account_id: str = Field(..., description="The account_id the generated booked cost entry will be booked at. "
                                             "It is an integer between 0 and 3999 formatted as string")

    account_id_validator = account_id_range_validator("account_id", 0, 3999)

    # Not used in settings, set by service applying the command cost entry (e.g. access service)
    source: Optional[str] = Field(default=None, description="Source of cost entry (usually service name)")
    source_id: Optional[str] = Field(default=None, description="Id to get information about cost entry in source "
                                                               "service")
    idempotency_key: Optional[str] = Field(default=None, description="Legacy field for idempotency key")

    @model_validator(mode="before")
    def data_validation(cls, values):
        if values.get("entry_type") in [CommandCostEntryTypes.static_cost, CommandCostEntryTypes.pending_static_cost,
                                        CommandCostEntryTypes.amount_validation, CommandCostEntryTypes.time_validation,
                                        CommandCostEntryTypes.payment]:
            if not isinstance(values.get("value"), int) or values.get("value") <= 0:
                raise ValueError("value must be an integer greater 0")
        elif values.get("entry_type") in [CommandCostEntryTypes.rate_change, CommandCostEntryTypes.rate_validation,
                                          CommandCostEntryTypes.partial_rate_change]:
            values["value"] = str(values.get("value"))
        if values.get("entry_type") not in [CommandCostEntryTypes.payment, CommandCostEntryTypes.cancel_remaining_costs,
                                            CommandCostEntryTypes.payoff_costs, CommandCostEntryTypes.free_timerange]:
            if values.get("group") is None:
                raise ValueError("group must be set")
        elif values.get("entry_type") in [CommandCostEntryTypes.cancel_remaining_costs,
                                          CommandCostEntryTypes.payoff_costs]:
            if values.get("group") is not None:
                raise ValueError("group must NOT be set")
        elif values.get("entry_type") == CommandCostEntryTypes.free_timerange:
            values["value"] = dict(entry_start=values.get("value", {}).get("entry_start"),
                                   entry_end=values.get("value", {}).get("entry_end"))
            if values.get("group") is None:
                raise ValueError("group must be set")
        return values


def cost_entry_validator(cost_entry):
    if cost_entry.entry_type in (CommandCostEntryTypes.partial_rate_change, CommandCostEntryTypes.payment,
                                 CommandCostEntryTypes.cancel_remaining_costs, CommandCostEntryTypes.payoff_costs):
        raise ValueError(f"{cost_entry.entry_type} is not allowed as cost entry")
    return cost_entry


CostEntryField = List[Annotated[CostEntryData, AfterValidator(cost_entry_validator)]]


class ShorttermLimitType(StrEnum):
    """
    Types of shortterm limits
    """
    no_limit = "no_limit"
    """no limit"""

    shortterm_count = "shortterm_count"
    """number of shortterm parkers in area is compared to limit"""

    total_count = "total_count"
    """number of all parkers in area are compared to limit"""


class ParkingArea(SettingsModel):
    """
    Defines a physical area of the whole parking lot by specifying the gates of types entry, exit, bidirectional or
    transit_barrier/transit_access_check. Also defines which gates allow shortterm parkers, the default cost entries
    for unknown parkers and pay-per-use parkers and limits for shortterm parkers in this area.
    This is the config for counters and shortterm parkers (registered and unregistered)
    """
    id: str = Field(..., description="Id of the parking area. UUID string in canonical textual representation")
    name: str = Field(..., description="Name for UIs", max_length=126)
    gates: List[str] = Field(..., description="Gates in this physical area (only types entry, exit, bidirectional or "
                                              "transit_barrier/transit_access_check)", min_length=1)
    shortterm_gates: List[str] = Field(..., description="Gates in this physical area (only types entry, exit, "
                                                        "bidirectional or transit_barrier/transit_access_check), where "
                                                        "shortterm parkers are allowed")
    default_cost_entries: CostEntryField = Field(..., description="Default cost entries added on entry "
                                                                  "in this parking area for unknown parkers")
    pay_per_use_cost_entries: Optional[CostEntryField] = \
        Field(default=None, description="Default cost entries added on entry in this parking area for pay-per-use "
                                        "parkers. None is fallback to default_cost_entries")
    shortterm_limit_type: ShorttermLimitType = Field(..., description="Type of shortterm limit")
    shortterm_limit: int = Field(..., description="Limit value in this area. Compared to value specified in "
                                                  "shortterm_limit_type", ge=0)
    time_based_shortterm_limit: Optional[str] = \
        Field(default=None, description="Overrules shortterm_limit if given. This is a yaml string defining different "
                                        "shortterm limits for different times")
    number_of_parking_spaces: Optional[int] = Field(default=None, description="Number of parking spaces in this area. "
                                                                              "None if not specified", ge=0)

    id_validator = uuid_validator("id")

    @model_validator(mode="after")
    def validate_shortterm_gates_in_gates(self):
        """
        check if all shortterm gates are in gates
        """
        if not all(gate in self.gates for gate in self.shortterm_gates):
            raise ValueError("All shortterm gates must be in gates")
        if len(self.shortterm_gates) == 0 and len(self.default_cost_entries) > 0:
            raise ValueError("If no shortterm gates are defined, no default cost entries are allowed")
        return self


class Parksetting(SettingsModel):
    """
    Defines the terms and conditions for parking sessions of registered users. This is not used for registered
    shortterm parkers.
    """
    id: str = Field(..., description="Id of the parksetting. UUID string in canonical textual representation")
    name: str = Field(..., description="Name for UIs", max_length=126)
    gates: List[str] = Field(..., description="Gates this parksetting is valid for (only types entry, exit, "
                                              "bidirectional or transit_barrier/transit_access_check", min_length=1)
    default_cost_entries: CostEntryField = Field(..., description="Default cost entries added on entry. "
                                                                  "Empty list is free.")

    id_validator = uuid_validator("id")


class CostGroup(SettingsModel):
    """
    Definition of cost group, specifying name, account_id and VAT rate.
    """
    id: str = Field(..., description="Id of the cost group. Scheme of this ID see "
                                     "https://youtrack.acc.si/youtrack/articles/DEV-A-1080#kostengruppen")
    name: str = Field(..., description="Name for UIs", max_length=126)
    account_id: str = Field(..., description="The account_id set in cost entries using this cost group (only costs)"
                                             "It is an integer between 0 and 3999 formatted as string")
    vat_rate: float = Field(..., ge=0.0, le=100.0, description="VAT rate of this cost group in %")

    account_id_validator = account_id_range_validator("account_id", 0, 999)

    @model_validator(mode="after")
    def data_validation(self):
        """
        check if cost group id prefix matches account_id range
        """
        # validator is run before root_validator, so int cast is safe
        if int(self.account_id) < 100 and not self.id.startswith("parking_"):
            raise ValueError("account_ids from 0 to 99 must have cost_group id parking_*")
        if 100 <= int(self.account_id) < 200 and not self.id.startswith("fine_"):
            raise ValueError("account_ids from 100 to 199 must have cost_group id fine_*")
        if 200 <= int(self.account_id) < 300 and not self.id.startswith("ev-charging_"):
            raise ValueError("account_ids from 200 to 299 must have cost_group id ev-charging_*")
        if 300 <= int(self.account_id) < 400 and not self.id.startswith("honest-payment_"):
            raise ValueError("account_ids from 300 to 399 must have cost_group id honest-payment_*")
        if 400 <= int(self.account_id) < 500 and not self.id.startswith("misc_"):
            raise ValueError("account_ids from 400 to 499 must have cost_group id misc_*")
        return self


class PrivacyTime(SettingsModel):
    """
    Model to define timespan as used in privacy settings. At least one of minutes, hours, days must > 0
    """
    minutes: int = Field(default=0, ge=0, description="Minutes of timespan. Must be >= 0 if given")
    hours: int = Field(default=0, ge=0, description="Hours of timespan. Must be >= 0 if given")
    days: int = Field(default=0, ge=0, description="Days of timespan. Must be >= 0 if given")

    @model_validator(mode="after")
    def data_validation(self):
        if self.minutes == 0 and self.hours == 0 and self.days == 0:
            raise ValueError("At least one of minutes, hours or days must be > 0")
        return self


class PrivacySettings(SettingsModel):
    """
    Settings for data anonymization in the parking system (on the device)
    """
    paid: PrivacyTime = Field(..., description="For parking sessions with all costs paid")
    unpaid: PrivacyTime = Field(..., description="For parking sessions with unpaid costs")
    registered_free: PrivacyTime = Field(..., description="For parking sessions with a free parking permission")
    pay_via_invoice: PrivacyTime = Field(..., description="For parking sessions which are paid later via invoice")
    open: PrivacyTime = Field(..., description="Unfinished parking sessions")
    free: PrivacyTime = Field(..., description="Data of customers which did not really use the parking lot "
                                               "(no costs, no registration)")
    honest_payment: PrivacyTime = Field(..., description="Data of honest payments")
    erroneous: PrivacyTime = Field(..., description="Data of erroneous incidents, e.g. unknown vehicle at the exit")
    rejected: PrivacyTime = Field(..., description="Data of rejections, where someone could not enter the parking lot")


class AddressModel(SettingsModel):
    """
    Address data, e.g. for billing address
    """
    name: str = Field(..., description="Name of entity at this address, e.g. provider name")

    street: str = Field(..., description="Street name")
    number: str = Field(..., description="House number")
    floor: Optional[str] = Field(default=None, description="Floor number or name")
    door: Optional[str] = Field(default=None, description="Door number or name")
    supplements: Optional[str] = Field(default=None, description="Additional description of address")

    city: str = Field(..., description="City")
    zip_code: str = Field(..., description="Zip code")
    state: Optional[str] = Field(default=None, description="State")
    country: str = Field(..., description="Country")


class BillingSettings(SettingsModel):
    """
    Settings for billing
    """
    vat_id: str = Field(..., description="VAT id of payment receiver")
    billing_address: AddressModel = Field(..., description="Address payment receiver")


class Location(AddressModel):
    """
    Address data with coordinates, e.g. for garage location
    """
    longitude: float = Field(..., description="Longitude of location")
    latitude: float = Field(..., description="Latitude of location")


class SupportSettings(SettingsModel):
    """
    Contact data for support cases (e.g. shown at signs, displays or end user UIs).
    Support means questions to contracts or complaints against costs and NOT emergency calls
    (cannot enter door, cannot pay, ...)
    """
    name: str = Field(..., description="Name of the person/entity responsible for support, e.g. provider")
    address: Optional[AddressModel] = Field(default=None, description="Address for written support letters")
    phone_number: Optional[constr(min_length=3, max_length=32, pattern=r"^\+[0-9 ]+$")] = \
        Field(default=None, description="Phone number for support (not emergency calls)")
    email_address: Optional[constr(min_length=3, max_length=254,
                                   pattern=r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$")] = \
        Field(default=None, description="Email address for support")
    website: Optional[constr(min_length=3, max_length=1022)] = Field(default=None, description="Website for support")

    @model_validator(mode="after")
    def data_validation(self):
        if (self.address or self.phone_number or self.email_address or self.website) is None:
            raise ValueError("At least one of address, phone_number, email_address or website must be given")
        return self


class GarageModes(StrEnum):
    """
    Garage modes
    """
    barrier = "barrier"
    """Standard barrier mode for commercial parking lots (barriers at entries and exits)"""

    freeflow = "freeflow"
    """Standard freeflow mode for commercial parking lots (no barriers at all)"""

    freeflow_surveillance = "freeflow_surveillance"
    """Surveillance only freeflow mode. No payments. Just observation of parking lot"""

    barrier_bypass = "barrier_bypass"
    """Bypass barrier mode, where unregistered parkers are handled by a third-party system"""


class GarageSettings(SettingsModel):
    """
    General garage settings defining general behavior of parking lot. Contains mode and flags for features which are
    enabled/disabled depending on the mode
    """
    mode: GarageModes = Field(..., description="General parking lot mode, e.g. freeflow. This field is more for "
                                               "information in UI and setup than for real technical use.")
    honest_payment_enabled: bool = Field(default=False, description="If honest payments are allowed")
    enforcement_enabled: bool = Field(default=False, description="If unpaid parking sessions can be enforced")
    payment_possible: bool = Field(default=False, description="If any parking system payments (Kiosk, PayApp,...) are "
                                                              "possible")

    @model_validator(mode="before")
    def data_validation(cls, values):
        if not isinstance(values, dict):
            return values
        if values["mode"] == GarageModes.freeflow:
            values["honest_payment_enabled"] = True
            values["enforcement_enabled"] = True
            values["payment_possible"] = True
        elif values["mode"] == GarageModes.barrier:
            values["honest_payment_enabled"] = False
            values["enforcement_enabled"] = False
            values["payment_possible"] = True
        elif values["mode"] == GarageModes.freeflow_surveillance:
            values["honest_payment_enabled"] = False
            values["enforcement_enabled"] = False
            values["payment_possible"] = False
        elif values["mode"] == GarageModes.barrier_bypass:
            values["honest_payment_enabled"] = False
            values["enforcement_enabled"] = False
            values["payment_possible"] = False
        return values


class GarageName(SettingsModel):
    """
    Names of Garage (directly set in Garage UI / iam)
    """
    name: str = Field(..., description="Name of Garage for UIs and end users", max_length=126)
    technical_name: constr(min_length=3, max_length=63, pattern=r"^[a-z0-9\-]+$") = \
        Field(..., description="Technical/internal name of garage")
    slug: constr(min_length=3, max_length=126, pattern=r"^[a-zA-Z0-9\-_]+$") = \
        Field(..., description="Shortname of Garage used in URLs")


class Urls(SettingsModel):
    """
    Collection of URLs used by various services (mainly for generating QRs or displaying links to user)
    """
    payapp_domain: str = Field(..., description="Domain of Arivo PayApp")
    payapp_short_url: str = Field(..., description="Full URL for PayApp with garage set")
    receipt_domain: str = Field(..., description="Domain of Arivo receipt website")

    payapp_domain_validator = domain_validator("payapp_domain")
    receipt_domain_validator = domain_validator("receipt_domain")


Gates = Dict[str, Gate]
Rates = Dict[str, Rate]
ParkingAreas = Dict[str, ParkingArea]
Parksettings = Dict[str, Parksetting]
CostGroups = Dict[str, CostGroup]



from decimal import Decimal

from settings_models._combat import SettingsModel
from settings_models.serialization import parse_setting, dump_setting, parse_setting_from_obj
from tests.utils import TestCase


class TestSerialization(TestCase):
    def test_unicode(self):
        setting_str = '{"name": "Testgarage ÄÖÜß", "technical_name": "at-8-test-garage", "slug": "testgarage"}'
        res = parse_setting("common/garage_name", setting_str)
        s = dump_setting(res)
        self.assertNotIn("Ä", s.decode())
        self.assertNotIn("Ö", s.decode())
        self.assertNotIn("Ü", s.decode())
        self.assertNotIn("ß", s.decode())
        res = parse_setting("common/garage_name", s)
        self.assertEqual(res.name, "Testgarage ÄÖÜß")

    def test_unicode_emoji(self):
        setting_str = '{"name": "Testgarage 😀", "technical_name": "at-8-test-garage", "slug": "testgarage"}'
        res = parse_setting("common/garage_name", setting_str)
        s = dump_setting(res)
        self.assertNotIn("😀", s.decode())
        res = parse_setting("common/garage_name", s)
        self.assertEqual(res.name, "Testgarage 😀")

    def test_custom_type(self):
        class CustomType(SettingsModel):
            name: str
            slug: str

        gt = dump_setting({"name": "Testgarage", "slug": "testgarage", "technical_name": "test"})
        res = parse_setting("common/garage_name", gt)
        self.assert_result(res, gt)

        res = parse_setting("common/garage_name", gt, custom_type=CustomType)
        self.assert_not_result(res, gt)
        self.assert_result(res, {"name": "Testgarage", "slug": "testgarage"})

    def test_parse_obj(self):
        gt = {"name": "Testgarage", "slug": "testgarage", "technical_name": "test"}
        with self.assertRaises(ValueError) as e:
            parse_setting("common/garage_name", gt)
        self.assertEqual(str(e.exception), "common/garage_name requires json decodable value")
        res = parse_setting_from_obj("common/garage_name", gt)
        self.assert_result(res, gt)

    def test_not_json_str(self):
        gt = "not json"
        with self.assertRaises(ValueError) as e:
            parse_setting("common/garage_name", gt)
        self.assertEqual(str(e.exception), "common/garage_name requires json decodable value")

    def test_unknown_setting(self):
        gt = '{"name": "Testgarage", "slug": "testgarage", "technical_name": "test"}'
        with self.assertRaises(KeyError) as e:
            parse_setting("common/unknown", gt)
        self.assertEqual(str(e.exception), "'No model for setting common/unknown found'")

    def test_parse_decimal(self):
        setting_val = {"parking_default": {
            "id": "parking_default", "name": "Default", "vat_rate": Decimal("0.19"), "account_id": Decimal(20)}}
        res = parse_setting("common/cost_groups", dump_setting(setting_val))
        self.assert_not_result(res, setting_val)
        self.assert_result(res, {"parking_default": {
            "id": "parking_default", "name": "Default", "vat_rate": 0.19, "account_id": "20"}})

    def test_not_json_dumpable(self):
        class CustomStr:
            def __init__(self, val):
                self.val = val

        setting_val = {"name": CustomStr("Testgarage"), "slug": "testgarage", "technical_name": "test"}
        with self.assertRaises(TypeError):
            dump_setting(setting_val)



from enum import StrEnum
from typing import Optional, Dict, Union

from pydantic import model_validator

from settings_models._combat import SettingsModel, Field


class SignalType(StrEnum):
    permanently_closed = "permanently_closed"
    shortterm_full = "shortterm_full"
    open = "open"
    close_pulse = "close_pulse"
    present_raw = "present_raw"
    present_decision = "present_decision"
    traffic_light_green = "traffic_light_green"
    area_full = "area_full"
    parkinglot_full = "parkinglot_full"
    custom = "custom"

    @staticmethod
    def gate_related_types():
        return [SignalType.permanently_closed, SignalType.shortterm_full, SignalType.open, SignalType.close_pulse,
                SignalType.present_raw, SignalType.present_decision, SignalType.traffic_light_green]


class SignalDefinition(SettingsModel):
    """
    Settings for available signals
    """
    signal: str = Field(..., description="Technical name of signal")
    name: str = Field(..., description="Human readable name of signal")
    type: Union[SignalType, str] = Field(..., description="Type of signal defining its behavior")
    gate: Optional[str] = Field(default=None, description="Gate of signal for gate related signals")
    parking_area_id: Optional[str] = Field(default=None, description="Parking area of signal for gate related signals")

    @model_validator(mode="after")
    def data_validation(self):
        if (self.type == SignalType.parkinglot_full and
                (self.gate is not None or self.parking_area_id is not None)):
            raise ValueError("gate and parking_area_id must be None for parkinglot_full signal")
        elif self.type == SignalType.parkinglot_full and self.signal != "parkinglot_full":
            raise ValueError("only signal parkinglot_full can be of type parkinglot_full")
        elif (self.type == SignalType.area_full and
              (self.gate is not None or self.parking_area_id is None)):
            raise ValueError("parking_area_id must not be None and gate must be None for area_full signal")
        elif (self.type in SignalType.gate_related_types() and
              (self.gate is None or self.parking_area_id is not None)):
            raise ValueError("gate must not be None and parking_area_id must be None for gate related signal")
        return self


class InputType(StrEnum):
    presence_loop = "presence_loop"
    safety_loop = "safety_loop"
    presence_laserscanner = "presence_laserscanner"
    safety_laserscanner = "safety_laserscanner"
    presence_narrow = "presence_narrow"
    safety_narrow = "safety_narrow"
    physical_entry_button = "physical_entry_button"
    custom = "custom"


class InputDefinition(SettingsModel):
    """
    Settings for available inputs
    """
    input: str = Field(..., description="Technical name of input")
    name: str = Field(..., description="Human readable name of input")
    type: Union[InputType, str] = Field(..., description="Type of input defining its behavior and usage")
    gate: Optional[str] = Field(default=None, description="Gate of input for all inputs except custom, where it's "
                                                          "optional")
    active: bool = Field(..., description="Whether input is active. False for preconfigured or disabled inputs")

    @model_validator(mode="after")
    def data_validation(self):
        if self.type != InputType.custom and self.gate is None:
            raise ValueError("gate must not be None for inputs except of type custom")
        return self


SignalDefinitions = Dict[str, SignalDefinition]
InputDefinitions = Dict[str, InputDefinition]



import warnings
from datetime import datetime
from unittest.mock import patch

from pydantic import ValidationError

from settings_models.serialization import dump_setting, parse_setting
from settings_models.settings.common import Parksettings
from settings_models.settings.io import SignalType, InputType
from tests.utils import TestCase


class ValidationTests(TestCase):
    def test_gates_success(self):
        gt = dump_setting({"gate1": {"gate": "gate1", "name": "EF", "type": "entry", "opposite_gate": "gate2",
                                     "is_wide": None},
                           "gate2": {"gate": "gate2", "name": "AF", "type": "exit", "opposite_gate": "gate1",
                                     "is_wide": None}})
        res = parse_setting("common/gates", gt)
        self.assertEqual(dump_setting(res), gt)

    def test_gates_success_unknown_type(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/gates",
                          dump_setting({"gate1": {"gate": "gate1", "name": "EF", "type": "wrong"},
                                        "gate2": {"gate": "gate2", "name": "AF", "type": "exit"}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("[type=enum, input_value='wrong', input_type=str]", str(e.exception))

    def test_gates_success_key_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("common/gates",
                          dump_setting({"gate1": {"gate": "gate3", "name": "EF", "type": "entry"},
                                        "gate2": {"gate": "gate2", "name": "AF", "type": "exit"}}))
        self.assertEqual("key must match field gate", str(e.exception))

    def test_gates_invalid_key(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/gates",
                          dump_setting({"gateß": {"gate": "gateß", "name": "EF", "type": "entry"}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("String should match pattern", str(e.exception))

    def test_opposite_gate_error(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("common/gates",
                          dump_setting({"gate1": {"gate": "gate1", "name": "EF", "type": "entry",
                                                  "opposite_gate": "gate2"},
                                        "gate2": {"gate": "gate2", "name": "AF", "type": "exit"}}))
        self.assertIn("gate gate1 and gate2 must reference each other", str(e.exception))

    def test_gates_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/gates", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_parksettings_success(self):
        self.maxDiff = None
        gt = dump_setting({
            "896150c9-5616-4406-a544-84c3824cfea9": {
                "id": "896150c9-5616-4406-a544-84c3824cfea9", "name": "Dauerparker mit variablen Kosten",
                "gates": ["gate1"], "default_cost_entries": []},
            "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179", "name": "Dauerparker ohne Kosten",
                "gates": ["gate1"],
                "default_cost_entries": [{"entry_type": "static_cost", "value": 123,
                                          "group": "parking_default", "account_id": "0",
                                          "source": None, "source_id": None, "idempotency_key": None}]}
        })
        res = parse_setting("common/parksettings2", gt)
        self.assert_result(dump_setting(res), gt)

    def test_parksettings_gates_no_empty_list(self):
        self.maxDiff = None
        with self.assertRaises(ValidationError) as e:
            gt = dump_setting({
                "896150c9-5616-4406-a544-84c3824cfea9": {
                    "id": "896150c9-5616-4406-a544-84c3824cfea9", "name": "Dauerparker mit variablen Kosten",
                    "gates": [], "default_cost_entries": []}})
            parse_setting("common/parksettings2", gt)
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("List should have at least 1 item after validation, not 0", str(e.exception))

    def test_parksettings_not_a_uuid(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                    "id": "123",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": []}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("must be a uuid string", str(e.exception))

    def test_parksettings_id_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "9999a7e2-be78-488e-8d1a-1486f7da9999": {
                    "id": "0000a7e2-be78-488e-8d1a-1486f7da0000",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": []}}))
        self.assertEqual("key must match field id", str(e.exception))

        with self.assertRaises(ValueError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "123": {
                    "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": []}}))
        self.assertIn("key must match field id", str(e.exception))

    def test_parksettings_invalid_default_cost_entry_type(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "0000a7e2-be78-488e-8d1a-1486f7da0000": {
                    "id": "0000a7e2-be78-488e-8d1a-1486f7da0000",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": [
                        {"entry_type": "partial_rate_change", "value": 123,
                         "group": "parking_default", "account_id": "0", "source": "test",
                         "source_id": "1", "idempotency_key": "1234"}
                    ]}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("is not allowed as cost entry", str(e.exception))

    def test_parksettings_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_cost_entry_invalid_type(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                    "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": [
                        {"entry_type": "static_cost123", "value": 123,
                         "group": "parking_default", "account_id": "0", "source": "test",
                         "source_id": "1", "idempotency_key": "1234"}
                    ]}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("[type=enum, input_value='static_cost123', input_type=str]", str(e.exception))

    def test_cost_entry_invalid_value(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                    "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": [
                        {"entry_type": "static_cost", "value": "NaN",
                         "group": "parking_default", "account_id": "0", "source": "test",
                         "source_id": "1", "idempotency_key": "1234"}
                    ]}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("value must be an integer greater 0", str(e.exception))

    def test_cost_entry_group_missing(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                    "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": [
                        {"entry_type": "static_cost", "value": 123,
                         "account_id": "0", "source": "test",
                         "source_id": "1", "idempotency_key": "1234"}
                    ]}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("group must be set", str(e.exception))

    def test_cost_entry_group_unexpectedly_set(self):
        # entry_type cancel_remaining_costs/payoff_costs
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                    "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": [
                        {"entry_type": "cancel_remaining_costs", "value": 123,
                         "account_id": "0", "source": "test", "group": "parking_default",
                         "source_id": "1", "idempotency_key": "1234"}
                    ]}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("group must NOT be set", str(e.exception))

    def test_cost_entry_free_timerange(self):
        # entry_type free timerange
        res = parse_setting("common/parksettings2", dump_setting({
            "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                "name": "Dauerparker ohne Kosten",
                "gates": ["gate1"],
                "default_cost_entries": [
                    {"entry_type": "free_timerange", "value": {"entry_start": "2021-01-01 00:00:00"},
                     "account_id": "0", "source": "test", "group": "parking_default",
                     "source_id": "1", "idempotency_key": "1234"}
                ]}}))
        self.assertDictEqual(dict(entry_start=datetime(2021, 1, 1, 0, 0, 0),
                                  entry_end=None),
                             res["12e8a7e2-be78-488e-8d1a-1486f7da2179"].default_cost_entries[0].value)
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parksettings2", dump_setting({
                "12e8a7e2-be78-488e-8d1a-1486f7da2179": {
                    "id": "12e8a7e2-be78-488e-8d1a-1486f7da2179",
                    "name": "Dauerparker ohne Kosten",
                    "gates": ["gate1"],
                    "default_cost_entries": [
                        {"entry_type": "free_timerange", "value": {"entry_start": "2021-01-01 00:00:00"},
                         "account_id": "0", "source": "test", "source_id": "1", "idempotency_key": "1234"}
                    ]}}))
        self.assertIn("group must be set", str(e.exception))

    def test_privacy_success(self):
        gt = dump_setting({"paid": {"days": 1, "hours": 0, "minutes": 0},
                           "unpaid": {"days": 90, "hours": 0, "minutes": 0},
                           "registered_free": {"days": 30, "hours": 0, "minutes": 0},
                           "pay_via_invoice": {"days": 30, "hours": 0, "minutes": 0},
                           "open": {"days": 90, "hours": 0, "minutes": 0},
                           "free": {"days": 0, "hours": 0, "minutes": 5},
                           "honest_payment": {"days": 90, "hours": 0, "minutes": 0},
                           "erroneous": {"days": 30, "hours": 0, "minutes": 0},
                           "rejected": {"days": 7, "hours": 0, "minutes": 0}})
        res = parse_setting("common/privacy_settings", gt)
        self.assert_result(res, gt)

    def test_privacy_paid_missing(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/privacy_settings",
                          dump_setting({"unpaid": {"days": 90, "hours": 0, "minutes": 0},
                                        "registered_free": {"days": 30, "hours": 0, "minutes": 0},
                                        "pay_via_invoice": {"days": 30, "hours": 0, "minutes": 0},
                                        "open": {"days": 90, "hours": 0, "minutes": 0},
                                        "free": {"days": 0, "hours": 0, "minutes": 5},
                                        "honest_payment": {"days": 90, "hours": 0, "minutes": 0},
                                        "erroneous": {"days": 30, "hours": 0, "minutes": 0},
                                        "rejected": {"days": 7, "hours": 0, "minutes": 0}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Field required", str(e.exception))

    def test_privacy_require_gt_0(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/privacy_settings",
                          dump_setting({"paid": {"hours": 0},
                                        "unpaid": {"days": 90, "hours": 0, "minutes": 0},
                                        "registered_free": {"days": 30, "hours": 0, "minutes": 0},
                                        "pay_via_invoice": {"days": 30, "hours": 0, "minutes": 0},
                                        "open": {"days": 90, "hours": 0, "minutes": 0},
                                        "free": {"days": 0, "hours": 0, "minutes": 5},
                                        "honest_payment": {"days": 90, "hours": 0, "minutes": 0},
                                        "erroneous": {"days": 30, "hours": 0, "minutes": 0},
                                        "rejected": {"days": 7, "hours": 0, "minutes": 0}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("At least one of minutes, hours or days must be > 0", str(e.exception))

    def test_privacy_negative_value(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/privacy_settings",
                          dump_setting({"paid": {"hours": 10},
                                        "unpaid": {"days": -90, "hours": 0, "minutes": 0},
                                        "registered_free": {"days": 30, "hours": 0, "minutes": 0},
                                        "pay_via_invoice": {"days": 30, "hours": 0, "minutes": 0},
                                        "open": {"days": 90, "hours": 0, "minutes": 0},
                                        "free": {"days": 0, "hours": 0, "minutes": 5},
                                        "honest_payment": {"days": 90, "hours": 0, "minutes": 0},
                                        "erroneous": {"days": 30, "hours": 0, "minutes": 0},
                                        "rejected": {"days": 7, "hours": 0, "minutes": 0}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be greater than or equal to 0", str(e.exception))

    def test_privacy_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/privacy_settings", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_cost_groups_success(self):
        gt = dump_setting({"honest-payment_default": {"account_id": "300", "id": "honest-payment_default",
                                                      "name": "Ehrliche Zahlung", "vat_rate": 20},
                           "parking_default": {"account_id": "0", "id": "parking_default", "name": "Parkkosten",
                                               "vat_rate": 20}})
        res = parse_setting("common/cost_groups", gt)
        self.assert_result(res, gt)

    def test_cost_groups_account_id_oob(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"honest-payment_default": {"account_id": 1234, "id": "honest-payment_default",
                                            "name": "Ehrliche Zahlung", "vat_rate": 20},
                 "parking_default": {"account_id": "0", "id": "parking_default",
                                     "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn('account_id must be a value between "0" and "999"', str(e.exception))

    def test_cost_groups_vat_over_100(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"honest-payment_default": {"account_id": "300", "id": "honest-payment_default",
                                            "name": "Ehrliche Zahlung", "vat_rate": 20},
                 "parking_default": {"account_id": "0", "id": "parking_default",
                                     "name": "Parkkosten", "vat_rate": 200}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("[type=less_than_equal, input_value=200, input_type=int]", str(e.exception))

    def test_cost_groups_account_id_nan(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"honest-payment_default": {"account_id": "NaN", "id": "honest-payment_default",
                                            "name": "Ehrliche Zahlung", "vat_rate": 20},
                 "parking_default": {"account_id": "0", "id": "parking_default",
                                     "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("account_id must be an integer as string", str(e.exception))

    def test_cost_groups_account_id_ranges(self):
        # 0-99 parking_*
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"asdf_default": {"account_id": "50", "id": "asdf_default", "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("account_ids from 0 to 99 must have cost_group id parking_*", str(e.exception))

        # 100-199 fine_*
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"asdf_default": {"account_id": "150", "id": "asdf_default", "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("account_ids from 100 to 199 must have cost_group id fine_*", str(e.exception))

        # 200-299 ev-charging_*
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"asdf_default": {"account_id": "250", "id": "asdf_default", "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("account_ids from 200 to 299 must have cost_group id ev-charging_*", str(e.exception))

        # 300-399 honest-payment_*
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"asdf_default": {"account_id": "350", "id": "asdf_default", "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("account_ids from 300 to 399 must have cost_group id honest-payment_*", str(e.exception))

        # 400-499 misc_*
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", dump_setting(
                {"asdf_default": {"account_id": "450", "id": "asdf_default", "name": "Parkkosten", "vat_rate": 20}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("account_ids from 400 to 499 must have cost_group id misc_*", str(e.exception))

        # 500+ any
        parse_setting("common/cost_groups", dump_setting(
            {"asdf_default": {"account_id": "550", "id": "asdf_default", "name": "Parkkosten", "vat_rate": 20}}))

    def test_cost_groups_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/cost_groups", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_rates_success(self):
        gt = {"7a50fafd-3426-42c8-bf3c-ca708638c327": {
            "id": "7a50fafd-3426-42c8-bf3c-ca708638c327", "name": "Kurzparktarif",
            "rate_yaml": "!RateMatrix\\nconstraints: []\\ntables:\\n  - !RateTable\\n     \
                      active_times: !OverallPeriod \\n      exception_periods: []\\n       \
                      valid_periods: []\\n    constraints: []\\n     \
                      id: 186ad735-6bcf-4c0c-9811-32bc8f502e17\\n    name: 3.30€ / Stunde\\n     \
                      line_collections:\\n      - !RateLineCollection\\n        grace_period: 3m\\n         \
                      max_value: 2500\\n        reset_duration: 1d\\n        lines:\\n           \
                      - !RateLine\\n            value: 330\\n            increment_period: 1h\\n"}}
        res = parse_setting("common/rates", dump_setting(gt))
        self.assert_result(gt, res)

    def test_rates_id_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("common/rates", dump_setting({"7a50fafd-3426-42c8-bf3c-ca708638c327": {
                "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "name": "Kurzparktarif",
                "rate_yaml": "!RateMatrix\\nconstraints: []\\ntables:\\n  - !RateTable\\n     \
                      active_times: !OverallPeriod \\n      exception_periods: []\\n       \
                      valid_periods: []\\n    constraints: []\\n     \
                      id: 186ad735-6bcf-4c0c-9811-32bc8f502e17\\n    name: 3.30€ / Stunde\\n     \
                      line_collections:\\n      - !RateLineCollection\\n        grace_period: 3m\\n         \
                      max_value: 2500\\n        reset_duration: 1d\\n        lines:\\n           \
                      - !RateLine\\n            value: 330\\n            increment_period: 1h\\n"}}))
        self.assertEqual("key must match field id", str(e.exception))

    def test_rates_not_a_uuid(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/rates", dump_setting({"123": {
                "id": "123", "name": "Kurzparktarif",
                "rate_yaml": "!RateMatrix\\nconstraints: []\\ntables:\\n  - !RateTable\\n     \
                     active_times: !OverallPeriod \\n      exception_periods: []\\n       \
                     valid_periods: []\\n    constraints: []\\n     \
                     id: 186ad735-6bcf-4c0c-9811-32bc8f502e17\\n    name: 3.30€ / Stunde\\n     \
                     line_collections:\\n      - !RateLineCollection\\n        grace_period: 3m\\n         \
                     max_value: 2500\\n        reset_duration: 1d\\n        lines:\\n           \
                     - !RateLine\\n            value: 330\\n            increment_period: 1h\\n"}}))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("must be a uuid string", str(e.exception))

    def test_rates_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/rates", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_parking_areas_success(self):
        self.maxDiff = None
        gt = dump_setting({"0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
            "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                [{"entry_type": "rate_change", "group": "parking_default",
                  "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0", "source": None, "source_id": None,
                  "idempotency_key": None}],
            "pay_per_use_cost_entries":
                [{"entry_type": "rate_change", "group": "parking_default",
                  "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0", "source": None, "source_id": None,
                  "idempotency_key": None}],
            "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": ["gate1"],
            "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
        })
        res = parse_setting("common/parking_areas2", gt)
        self.assert_result(dump_setting(res), gt)

    def test_parking_areas_no_cost_entries_without_shortterm(self):
        self.maxDiff = None
        with self.assertRaises(ValidationError) as e:
            gt = dump_setting({"0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                    [{"entry_type": "rate_change", "group": "parking_default",
                      "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                "pay_per_use_cost_entries":
                    [{"entry_type": "rate_change", "group": "parking_default",
                      "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": [],
                "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            })
            parse_setting("common/parking_areas2", gt)
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("If no shortterm gates are defined, no default cost entries are allowed", str(e.exception))

    def test_parking_areas_gates_no_empty_list(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                        [{"entry_type": "rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": [], "name": "Gesamter Parkplatz", "shortterm_gates": [],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("List should have at least 1 item after validation, not 0", str(e.exception))

    def test_parking_areas_not_a_uuid(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                    "id": "123", "default_cost_entries":
                        [{"entry_type": "rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": [],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("must be a uuid string", str(e.exception))

    def test_parking_areas_id_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "7a50fafd-3426-42c8-bf3c-ca708638c327": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                        [{"entry_type": "rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": ["gate1"],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual("key must match field id", str(e.exception))

        with self.assertRaises(ValueError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "123": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                        [{"entry_type": "rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": ["gate1"],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertIn("key must match field id", str(e.exception))

    def test_parking_areas_invalid_default_cost_entry_type(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                        [{"entry_type": "partial_rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": [],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("is not allowed as cost entry", str(e.exception))

    def test_parking_areas_invalid_pay_per_use_cost_entry_type(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries": [],
                    "pay_per_use_cost_entries":
                        [{"entry_type": "partial_rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": [],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("is not allowed as cost entry", str(e.exception))

    def test_parking_areas_shortterm_gates_not_in_gates(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                        [{"entry_type": "rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": ["gate1", "gate2"],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual(len(e.exception.errors()), 1)

        self.assertIn("All shortterm gates must be in gates", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", dump_setting({
                "0eec2445-15c3-4af8-b362-aeb1f278e3ac": {
                    "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", "default_cost_entries":
                        [{"entry_type": "rate_change", "group": "parking_default",
                          "value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0"}],
                    "gates": ["gate1"], "name": "Gesamter Parkplatz", "shortterm_gates": ["gate2"],
                    "shortterm_limit_type": "no_limit", "shortterm_limit": 0}
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("All shortterm gates must be in gates", str(e.exception))

    def test_parking_areas_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/parking_areas2", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_timezone(self):
        parse_setting("common/timezone", dump_setting("Europe/Berlin"))
        with self.assertRaises(ValueError) as e:
            parse_setting("common/timezone", dump_setting("Europe/Graz"))
        self.assertEqual("timezone Europe/Graz unknown", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/timezone", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid string", str(e.exception))

    def test_language(self):
        parse_setting("common/language", dump_setting("en"))
        with self.assertRaises(ValueError) as e:
            parse_setting("common/language", dump_setting("english"))
        self.assertEqual("english is no valid language code", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/language", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid string", str(e.exception))

    def test_currency(self):
        parse_setting("common/currency", dump_setting("EUR"))
        with self.assertRaises(ValueError) as e:
            parse_setting("common/currency", dump_setting("euros"))
        self.assertEqual("Currency must be 3 uppercase letters", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/currency", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid string", str(e.exception))

    def test_location_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/location", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_billing_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/billing", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_support_success(self):
        gt = dump_setting({
            "name": "Test", "address": {
                "name": "Garage", "street": "Garagestreet", "number": "12c",
                "floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz",
                "zip_code": "8020", "state": "Styria", "country": "Österreich"},
            "phone_number": "+43 123 456 789", "email_address": "ines@arivo.co", "website": "https://arivo.co/support"
        })
        res = parse_setting("common/support", gt)
        self.assert_result(res, gt)

    def test_support_not_given(self):
        gt = dump_setting({"name": "Test", "phone_number": "+43 123 456 789"})
        res = parse_setting("common/support", gt)
        self.assert_result(res, gt)
        gt = dump_setting({"name": "Test", "website": "https://arivo.co"})
        res = parse_setting("common/support", gt)
        self.assert_result(res, gt)

        with self.assertRaises(ValidationError) as e:
            gt = dump_setting({"name": "Test"})
            res = parse_setting("common/support", gt)
            self.assert_result(res, gt)
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("At least one of address, phone_number, email_address or website must be given", str(e.exception))

    def test_support_invalid_email(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/support", dump_setting({
                "name": "Test", "address": {
                    "name": "Garage", "street": "Garagestreet", "number": "12c",
                    "floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz",
                    "zip_code": "8020", "state": "Styria", "country": "Österreich"},
                "phone_number": "+43 123 456 789", "email_address": "my-great-email"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("String should match pattern", str(e.exception))

        with self.assertRaises(ValidationError) as e:
            parse_setting("common/support", dump_setting({
                "name": "Test", "address": {
                    "name": "Garage", "street": "Garagestreet", "number": "12c",
                    "floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz",
                    "zip_code": "8020", "state": "Styria", "country": "Österreich"},
                "phone_number": "+43 123 456 789", "email_address": ""
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("String should have at least 3 characters", str(e.exception))

    def test_support_invalid_phone_number(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/support", dump_setting({
                "name": "Test", "address": {
                    "name": "Garage", "street": "Garagestreet", "number": "12c",
                    "floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz",
                    "zip_code": "8020", "state": "Styria", "country": "Österreich"},
                "phone_number": "my phone number", "email_address": "ines@arivo.co"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("String should match pattern", str(e.exception))

        with self.assertRaises(ValidationError) as e:
            parse_setting("common/support", dump_setting({
                "name": "Test", "address": {
                    "name": "Garage", "street": "Garagestreet", "number": "12c",
                    "floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz",
                    "zip_code": "8020", "state": "Styria", "country": "Österreich"},
                "phone_number": "", "email_address": "ines@arivo.co"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("String should have at least 3 characters", str(e.exception))

    def test_support_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/support", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_garage_name(self):
        gt = dump_setting({"name": "Test", "slug": "test_123ABC", "technical_name": "test"})
        res = parse_setting("common/garage_name", gt)
        self.assert_result(dump_setting(res), gt)

        with self.assertRaises(ValidationError) as e:
            parse_setting("common/garage_name", dump_setting({
                "name": "Test", "slug": "test", "technical_name": "TEST"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertEqual(("technical_name",), e.exception.errors()[0]["loc"])
        self.assertIn("String should match pattern", str(e.exception))

        with self.assertRaises(ValidationError) as e:
            parse_setting("common/garage_name", dump_setting({
                "name": "test", "slug": "test.123", "technical_name": "test"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertEqual(("slug",), e.exception.errors()[0]["loc"])
        self.assertIn("String should match pattern", str(e.exception))

        with self.assertRaises(ValidationError) as e:
            parse_setting("common/garage_name", dump_setting({
                "name": "test", "slug": "töst", "technical_name": "test"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertEqual(("slug",), e.exception.errors()[0]["loc"])
        self.assertIn("String should match pattern", str(e.exception))

        with self.assertRaises(ValidationError) as e:
            parse_setting("common/garage_name", dump_setting({
                "name": "test", "slug": "test/123", "technical_name": "test"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertEqual(("slug",), e.exception.errors()[0]["loc"])
        self.assertIn("String should match pattern", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/garage_name", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_garage_settings_freeflow(self):
        expected_gt = dump_setting({"mode": "freeflow", "honest_payment_enabled": True,
                                    "enforcement_enabled": True, "payment_possible": True})
        gt = dump_setting({"mode": "freeflow"})
        res = parse_setting("common/garage_settings", gt)
        self.assert_result(dump_setting(res), expected_gt)

        gt = dump_setting({"mode": "freeflow", "honest_payment_enabled": False,
                           "enforcement_enabled": False, "payment_possible": False})
        res = parse_setting("common/garage_settings", gt)
        self.assert_result(dump_setting(res), expected_gt)

    def test_garage_settings_freeflow_surveillance(self):
        expected_gt = dump_setting({"mode": "freeflow_surveillance", "honest_payment_enabled": False,
                                    "enforcement_enabled": False, "payment_possible": False})

        gt = dump_setting({"mode": "freeflow_surveillance"})
        res = parse_setting("common/garage_settings", gt)
        self.assert_result(dump_setting(res), expected_gt)

        gt = dump_setting({"mode": "freeflow_surveillance", "honest_payment_enabled": True,
                           "enforcement_enabled": True, "payment_possible": True})
        res = parse_setting("common/garage_settings", gt)
        self.assert_result(dump_setting(res), expected_gt)

    def test_garage_settings_barrier(self):
        expected_gt = dump_setting({"mode": "barrier", "honest_payment_enabled": False,
                                    "enforcement_enabled": False, "payment_possible": True})

        gt = dump_setting({"mode": "barrier"})
        res = parse_setting("common/garage_settings", gt)
        self.assert_result(dump_setting(res), expected_gt)

        gt = dump_setting({"mode": "barrier", "honest_payment_enabled": True,
                           "enforcement_enabled": True, "payment_possible": False})
        res = parse_setting("common/garage_settings", gt)
        self.assert_result(dump_setting(res), expected_gt)

    def test_garage_settings_none(self):
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/garage_settings", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_gate_control_day_mode_invalid_format(self):
        # start
        with self.assertRaises(ValidationError) as e:
            parse_setting("gate_control/day_mode", dump_setting({
                "start": "24", "end": "00:00", "enabled": True
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Field must be string of form HH:MM", str(e.exception))

        # end
        with self.assertRaises(ValidationError) as e:
            parse_setting("gate_control/day_mode", dump_setting({
                "start": "00:00", "end": "24", "enabled": True
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Field must be string of form HH:MM", str(e.exception))

    def test_gate_control_day_mode_none(self):
        res = parse_setting("gate_control/day_mode", "null")
        self.assert_result("null", dump_setting(res))

    def test_gate_control_mode(self):
        res = parse_setting("gate_control/mode", "null")
        self.assert_result("null", dump_setting(res))

    def test_device_keys_pairing(self):
        gt = {"key": "1234567890", "created_at": "2000-01-01 02:00:00+01:00", "revision": 1}
        res = parse_setting("device_keys/pairing", dump_setting(gt))
        self.assert_not_result(res, gt)
        self.assertNotEqual(str(res.created_at), gt["created_at"])
        self.assertEqual(str(res.created_at), "2000-01-01 01:00:00")

        with self.assertRaises(ValidationError) as e:
            parse_setting("device_keys/pairing", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_urls(self):
        # payapp_domain
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/urls", dump_setting({
                "payapp_short_url": "https://payapp.arivo.co/test",
                "payapp_domain": "https://payapp.arivo.co/",
                "receipt_domain": "https://receipt.arivo.co"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Domains must not have a trailing /", str(e.exception))

        # receipt_domain
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/urls", dump_setting({
                "payapp_short_url": "https://payapp.arivo.co/test",
                "payapp_domain": "https://payapp.arivo.co",
                "receipt_domain": "https://receipt.arivo.co/"
            }))
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Domains must not have a trailing /", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("common/urls", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_custom_type_not_validated(self):
        # id does not match key, but custom type is used
        gt = {"9999a7e2-be78-488e-8d1a-1486f7da9999": {
            "id": "0000a7e2-be78-488e-8d1a-1486f7da0000",
            "name": "Dauerparker ohne Kosten",
            "gates": ["gate1"],
            "default_cost_entries": []}}
        res = parse_setting("common/parksettings2", dump_setting(gt), custom_type=Parksettings)
        self.assert_result(res, gt)

    def test_intercom(self):
        setting_str = '{"enabled": false, "phone_number": null}'
        res = parse_setting("intercom/basic_settings", setting_str)
        self.assert_result(setting_str, dump_setting(res))

        setting_str = '{"enabled": true, "phone_number": null}'
        with self.assertRaises(ValidationError) as e:
            parse_setting("intercom/basic_settings", setting_str)
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("phone_number must be set if intercom is enabled", str(e.exception))
        with self.assertRaises(ValidationError) as e:
            parse_setting("intercom/basic_settings", "null")
        self.assertEqual(len(e.exception.errors()), 1)
        self.assertIn("Input should be a valid dictionary", str(e.exception))

    def test_enforcement(self):
        setting_str = '{"strictness": 0, "payment_deadline_hours": 1, "enabled": true, ' \
                      '"ai_enabled": true, "last_edited": "2024-02-19T13:06:51.140075"}'
        res = parse_setting("enforcement/basic_settings", setting_str)
        self.assert_result(setting_str, dump_setting(res))

        setting_str = '{"strictness": 0, "payment_deadline_hours": 0, "enabled": true, ' \
                      '"ai_enabled": true, "last_edited": "2024-02-19T13:06:51.140075"}'
        with self.assertRaises(ValidationError) as e:
            parse_setting("enforcement/basic_settings", setting_str)
        self.assertIn("Input should be greater than 0", str(e.exception))

        setting_str = '{"strictness": 0, "payment_deadline_hours": 960, "enabled": true, ' \
                      '"ai_enabled": true, "last_edited": "2024-02-19T13:06:51.140075"}'  # 40 days * 24
        res = parse_setting("enforcement/basic_settings", setting_str)
        self.assert_result(setting_str, dump_setting(res))

        setting_str = '{"strictness": 0, "payment_deadline_hours": 961, "enabled": true, ' \
                      '"ai_enabled": true, "last_edited": "2024-02-19T13:06:51.140075"}'  # 40 days * 24 + 1
        with self.assertRaises(ValidationError) as e:
            parse_setting("enforcement/basic_settings", setting_str)
        self.assertIn("Input should be less than or equal to 960", str(e.exception))

        # test ai_enabled
        setting_str = '{"strictness": 0, "payment_deadline_hours": 1, "enabled": true, ' \
                      '"ai_enabled": false, "last_edited": "2027-11-11T11:11:11"}'
        setting = parse_setting("enforcement/basic_settings", setting_str)
        self.assertFalse(setting.ai_enabled)
        setting_str = '{"strictness": 0, "payment_deadline_hours": 1, "enabled": true, ' \
                      '"ai_enabled": true, "last_edited": "2027-11-11T11:11:11"}'
        setting = parse_setting("enforcement/basic_settings", setting_str)
        self.assertTrue(setting.ai_enabled)
        # if the value is not present, then the default value of activated should be used
        setting_str = '{"strictness": 0, "payment_deadline_hours": 1, "enabled": true, ' \
                      '"last_edited": "2027-11-11T11:11:11"}'
        setting = parse_setting("enforcement/basic_settings", setting_str)
        self.assertTrue(setting.ai_enabled)

        # test strictness is now optional and deprecated
        setting_str = '{"payment_deadline_hours": 1, "enabled": true, ' \
                      '"last_edited": "2027-11-11T11:11:11"}'
        setting = parse_setting("enforcement/basic_settings", setting_str)
        with warnings.catch_warnings(record=True) as w:
            strictness = setting.strictness
        self.assertEqual(strictness, 2)
        self.assertEqual(len(w), 1)
        self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        self.assertIn("The field `EnforcementSettings.strictness` is deprecated "
                      "and will be removed in a future version.", str(w[-1].message))

    def test_io_signals_correct_fields_gate_related(self):
        for t in [SignalType.permanently_closed, SignalType.shortterm_full, SignalType.open, SignalType.close_pulse,
                  SignalType.present_raw, SignalType.present_decision, SignalType.traffic_light_green]:
            setting_str = (
                    '{"entry-open": {"signal":"entry-open","gate":"entry","type":"' + t.value + '","name":"EF open"}}')
            res = parse_setting("io/signals", setting_str)
            self.assert_result(setting_str, dump_setting(res))

            setting_str = '{"entry-open": {"signal":"entry-open","type":"' + t.value + '","name":"EF open"}}'
            with self.assertRaises(ValueError, msg=t) as e:
                res = parse_setting("io/signals", setting_str)
            self.assertIn("gate must not be None and parking_area_id must be None for gate related signal",
                          str(e.exception), t)

            setting_str = ('{"entry-open":{"signal":"entry-open","gate":"entry","parking_area_id":"pa1","type":"'
                           + t.value + '","name":"EF open"}}')
            with self.assertRaises(ValueError, msg=t) as e:
                res = parse_setting("io/signals", setting_str)
            self.assertIn("gate must not be None and parking_area_id must be None for gate related signal",
                          str(e.exception), t)

    def test_io_signals_correct_fields_area_related(self):
        setting_str = \
            '{"pa1-area_full": {"signal":"pa1-area_full","parking_area_id":"pa1","type":"area_full","name":"PA1 full"}}'
        res = parse_setting("io/signals", setting_str)
        self.assert_result(setting_str, dump_setting(res))

        setting_str = '{"pa1-area_full": {"signal":"pa1-area_full","type":"area_full","name":"PA1 full"}}'
        with self.assertRaises(ValueError) as e:
            res = parse_setting("io/signals", setting_str)
        self.assertIn("parking_area_id must not be None and gate must be None for area_full signal", str(e.exception))

        setting_str = ('{"pa1-area_full": {"signal":"pa1-area_full","parking_area_id":"pa1","type":"area_full",'
                       '"name":"PA1 full","gate":"entry"}}')
        with self.assertRaises(ValueError) as e:
            res = parse_setting("io/signals", setting_str)
        self.assertIn("parking_area_id must not be None and gate must be None for area_full signal", str(e.exception))

    def test_io_signals_correct_fields_parkinglot_full(self):
        setting_str = '{"parkinglot_full": {"signal":"parkinglot_full","type":"parkinglot_full","name":"Full"}}'
        res = parse_setting("io/signals", setting_str)
        self.assert_result(setting_str, dump_setting(res))

        setting_str = \
            '{"parkinglot_full": {"signal":"parkinglot_full","type":"parkinglot_full","name":"Full","gate":"entry"}}'
        with self.assertRaises(ValueError) as e:
            res = parse_setting("io/signals", setting_str)
        self.assertIn("gate and parking_area_id must be None for parkinglot_full signal", str(e.exception))

        setting_str = ('{"parkinglot_full": {"signal":"parkinglot_full","type":"parkinglot_full","name":"Full",'
                       '"parking_area_id":"pa1"}}')
        with self.assertRaises(ValueError) as e:
            res = parse_setting("io/signals", setting_str)
        self.assertIn("gate and parking_area_id must be None for parkinglot_full signal", str(e.exception))

        setting_str = '{"test": {"signal":"test","type":"parkinglot_full","name":"Full"}}'
        with self.assertRaises(ValueError) as e:
            res = parse_setting("io/signals", setting_str)
        self.assertIn("only signal parkinglot_full can be of type parkinglot_full", str(e.exception))

    def test_io_signals_key_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("io/signals", '{"t": {"signal":"parkinglot_full","type":"parkinglot_full","name":"Full"}}')
        self.assertEqual("key must match field signal", str(e.exception))

    def test_io_inputs_correct_fields(self):
        for t in [InputType.presence_loop, InputType.presence_laserscanner, InputType.presence_narrow,
                  InputType.safety_loop, InputType.safety_laserscanner, InputType.safety_narrow]:
            setting_str = ('{"entry-t":{"input":"entry-t","type":"' + t.value +
                           '","gate":"entry","name":"EF CLoop","active":true}}')
            res = parse_setting("io/inputs", setting_str)
            self.assert_result(setting_str, dump_setting(res))

            setting_str = '{"entry-t":{"input":"entry-t","type":"' + t.value + '","name":"EF CLoop","active":true}}'
            with self.assertRaises(ValueError) as e:
                res = parse_setting("io/inputs", setting_str)
            self.assertIn("gate must not be None for inputs except of type custom", str(e.exception))

    def test_io_inputs_key_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("io/inputs", '{"t":{"input":"entry-safety_loop","type":"safety_loop","gate":"entry",'
                                       '"name":"EF CLoop","active":true}}')
        self.assertEqual(str(e.exception).count("\n"), 0)
        self.assertEqual("key must match field input", str(e.exception))

    def test_enforcement_none(self):
        res = parse_setting("enforcement/basic_settings", "null")
        self.assert_result("null", dump_setting(res))

    def test_assets_visual_key_does_not_match_field(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("assets/visuals", '{"x":{"type":"y","value":"z"}}')
        self.assertEqual(str(e.exception).count("\n"), 0)
        self.assertEqual("key must match field type", str(e.exception))

    def test_assets_visual_value(self):
        with self.assertRaises(ValueError) as e:
            parse_setting("assets/visuals", '{"logo":{"type":"logo","value":[]}}')
        self.assertIn("Field value must be a string (URL) for logo and marketing_image types", str(e.exception))

    @patch("settings_models.settings.assets.datetime")
    def test_assets_visual_auto_created(self, mock_datetime):
        t1 = datetime(2025, 1, 1, 12)
        t2 = datetime(2024, 12, 31, 12, 59, 59)
        mock_datetime.now.side_effect = [t1]
        res = parse_setting("assets/visuals", '{"x":{"type":"x","value":"y"},'
                                              '"a":{"type":"a","value":"b","created":"2024-12-31T12:59:59"}}')

        self.assertEqual(res["x"].created, t1)
        self.assertEqual(res["a"].created, t2)



from datetime import datetime
from typing import Optional

from pydantic import Field

from settings_models._combat import SettingsModel
from settings_models.validators import timezone_validator


class PairingKey(SettingsModel):
    """
    Pairing key data and metadata
    """
    key: str = Field(..., description="Pairing key")
    created_at: datetime = Field(..., description="Creation time of pairing key")
    revision: int = Field(..., ge=0, description="Revision of pairing key")

    created_at_validator = timezone_validator("created_at")



from datetime import datetime, UTC
from enum import StrEnum
from typing import Any, Annotated
from typing_extensions import Doc

from pydantic import model_validator

from settings_models._combat import SettingsModel, Field
from settings_models.validators import timezone_validator


def utcnow():
    return datetime.now(UTC).replace(tzinfo=None)


class AssetVisualType(StrEnum):
    logo = "logo"
    marketing_image = "marketing_image"  # legacy
    tariff_table = "tariff_table"
    payment_methods = "payment_methods"


class AssetVisualDefinition(SettingsModel):
    """
    Settings for available assets
    """
    type: AssetVisualType | str = Field(..., description="Type of asset defining its behavior")
    value: Any = Field(description="URL or config of the asset")
    created: datetime = Field(default_factory=utcnow, description="Creation time (UTC) of the asset")

    created_validator = timezone_validator("created")

    @model_validator(mode="after")
    def validate_value(self) -> "AssetVisualDefinition":
        if self.type in [AssetVisualType.logo, AssetVisualType.marketing_image] and not isinstance(self.value, str):
            raise ValueError("Field value must be a string (URL) for logo and marketing_image types")
        return self


AssetVisualDefinitions = Annotated[dict[AssetVisualType | str, AssetVisualDefinition], Doc(AssetVisualDefinition.__doc__)]



import re
import uuid
import warnings
from datetime import datetime

from dateutil.tz import UTC
from pydantic import field_validator
from pydantic_core.core_schema import ValidationInfo


def _uuid_validator(cls, data, values, **kwargs):
    try:
        uuid.UUID(data)
    except (TypeError, ValueError):
        raise ValueError("must be a uuid string")
    return data


def uuid_validator(field):
    return field_validator(field)(_uuid_validator)


def _hhmm_validator(cls, data, values, **kwargs):
    if re.search(r"^(0\d|1\d|2[0-3]):[0-5]\d$", data) is None:
        raise ValueError("Field must be string of form HH:MM")
    return data


def hhmm_validator(field):
    return field_validator(field)(_hhmm_validator)


def _timezone_validator(cls, dt: datetime, values, **kwargs):
    if dt and dt.tzinfo:
        return dt.astimezone(UTC).replace(tzinfo=None)
    else:
        return dt


def timezone_validator(field):
    return field_validator(field)(_timezone_validator)


def _domain_validator(cls, data, values, **kwargs):
    if data.endswith("/"):
        raise ValueError("Domains must not have a trailing /")
    return data


def domain_validator(field):
    return field_validator(field)(_domain_validator)


def _account_id_validator_wrapper(lower_limit, upper_limit):
    def _account_id_validator(cls, account_id, values, **kwargs):
        bad_range = False
        try:
            if not (lower_limit <= int(account_id) <= upper_limit):
                bad_range = True
        except ValueError:
            raise ValueError("account_id must be an integer as string")
        if bad_range:
            raise ValueError('account_id must be a value between "%s" and "%s"' % (lower_limit, upper_limit))
        return account_id

    return _account_id_validator


def account_id_range_validator(field, lower_limit=0, upper_limit=1000):
    return field_validator(field, mode="before")(
        _account_id_validator_wrapper(lower_limit, upper_limit))


def _deprecated_validator(cls, value, info: ValidationInfo, **kwargs):
    warnings.warn(
        f"The field `{cls.__name__}.{info.field_name}` is deprecated and will be removed in a future version.",
        DeprecationWarning,
        stacklevel=2
    )
    return value


def deprecated_validator(field):
    return field_validator(field, mode="before", check_fields=False)(_deprecated_validator)



import datetime
from types import FunctionType
from unittest.mock import patch

from settings_models import serialization
from settings_models._combat import SettingsModel
from settings_models.serialization import dump_setting, parse_setting
from tests.utils import TestCase


class ParsingTests(TestCase):
    def test_gates(self):
        setting_str = '{"gate1":{"gate":"gate1","name":"EF","type":"entry"},' \
                      '"gate2":{"gate":"gate2","name":"AF","type":"exit"}}'
        res = parse_setting("common/gates", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_rates(self):
        setting_str = ('{"7a50fafd-3426-42c8-bf3c-ca708638c327":{'
                       '"id":"7a50fafd-3426-42c8-bf3c-ca708638c327","name":"Kurzparktarif",'
                       '"rate_yaml":"!RateMatrix\\nconstraints: []\\ntables:\\n  - !RateTable\\n    '
                       'active_times: !OverallPeriod \\n      exception_periods: []\\n      '
                       'valid_periods: []\\n    constraints: []\\n    '
                       'id: 186ad735-6bcf-4c0c-9811-32bc8f502e17\\n    name: 3.30€ / Stunde\\n    '
                       'line_collections:\\n      - !RateLineCollection\\n        grace_period: 3m\\n        '
                       'max_value: 2500\\n        reset_duration: 1d\\n        lines:\\n          '
                       '- !RateLine\\n            value: 330\\n            increment_period: 1h\\n"}}')
        res = parse_setting("common/rates", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_parking_areas2(self):
        setting_str = ('{"0eec2445-15c3-4af8-b362-aeb1f278e3ac": {"default_cost_entries": '
                       '[{"entry_type": "rate_change", "group": "parking_default", '
                       '"value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0", "source": null, '
                       '"source_id": null, "idempotency_key": null}], '
                       '"gates": ["gate1"], "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", '
                       '"name": "Gesamter Parkplatz", "shortterm_gates": ["gate1"],'
                       '"shortterm_limit_type": "no_limit", "shortterm_limit": 0, "number_of_parking_spaces": 10}}')
        res = parse_setting("common/parking_areas2", setting_str)
        self.assert_result(setting_str, dump_setting(res))

        setting_str = ('{"0eec2445-15c3-4af8-b362-aeb1f278e3ac": {"default_cost_entries": '
                       '[{"entry_type": "rate_change", "group": "parking_default", '
                       '"value": "7a50fafd-3426-42c8-bf3c-ca708638c327", "account_id": "0", "source": null, '
                       '"source_id": null, "idempotency_key": null}], '
                       '"gates": ["gate1"], "id": "0eec2445-15c3-4af8-b362-aeb1f278e3ac", '
                       '"name": "Gesamter Parkplatz", "shortterm_gates": ["gate1"],'
                       '"shortterm_limit_type": "no_limit", "shortterm_limit": 0, "number_of_parking_spaces": 10,'
                       '"time_based_shortterm_limit": "some yaml"}}')
        res = parse_setting("common/parking_areas2", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_parksettings2(self):
        setting_str = ('{"654cb72d-545c-48e1-9de7-146528affb6c": {"name": "Dauerparker ohne Kosten", '
                       '"id": "654cb72d-545c-48e1-9de7-146528affb6c", "default_cost_entries": [], "gates": ["gate1"]}, '
                       '"70c3f84d-7408-487c-af0c-e41d8fe6c6ac": {"id": "70c3f84d-7408-487c-af0c-e41d8fe6c6ac", '
                       '"name": "Rabattparker", "gates": ["gate1", "gate2"], '
                       '"default_cost_entries": [{"entry_type": "rate_change", "group": "parking_default", '
                       '"value": "75dadac3-2362-4075-bae1-8a26e06f2d76", "account_id": "0", "source": null, '
                       '"source_id": null, "idempotency_key": null}]}}')
        res = parse_setting("common/parksettings2", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_privacy_settings(self):
        setting_str = ('{"paid": {"days": 1, "hours": 0, "minutes": 0}, '
                       '"unpaid": {"days": 90, "hours": 0, "minutes": 0}, '
                       '"registered_free": {"days": 30, "hours": 0, "minutes": 0}, '
                       '"pay_via_invoice": {"days": 30, "hours": 0, "minutes": 0}, '
                       '"open": {"days": 90, "hours": 0, "minutes": 0}, '
                       '"free": {"days": 0, "hours": 0, "minutes": 5}, '
                       '"honest_payment": {"days": 90, "hours": 0, "minutes": 0}, '
                       '"erroneous": {"days": 30, "hours": 0, "minutes": 0}, '
                       '"rejected": {"days": 7, "hours": 0, "minutes": 0}}')
        res = parse_setting("common/privacy_settings", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_cost_groups(self):
        setting_str = ('{"fine_no-ticket": {"account_id": "100", "id": "fine_no-ticket", '
                       '"name": "Strafe: Kein Ticket", "vat_rate": 0}, '
                       '"fine_parking-violation": {"account_id": "101", "id": "fine_parking-violation", '
                       '"name": "Strafe: Unerlaubter Stellplatz", "vat_rate": 0}, '
                       '"honest-payment_default": {"account_id": "300", "id": "honest-payment_default", '
                       '"name": "Ehrliche Zahlung", "vat_rate": 20}, '
                       '"parking_default": {"account_id": "0", "id": "parking_default", "name": '
                       '"Parkkosten", "vat_rate": 20}}')
        res = parse_setting("common/cost_groups", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_currency(self):
        setting_str = '"EUR"'
        res = parse_setting("common/currency", setting_str)
        self.assertEqual(setting_str.encode("utf8"), dump_setting(res))

    def test_language(self):
        setting_str = '"de"'
        res = parse_setting("common/language", setting_str)
        self.assertEqual(setting_str.encode("utf8"), dump_setting(res))

    def test_timezone(self):
        setting_str = '"Europe/Vienna"'
        res = parse_setting("common/timezone", setting_str)
        self.assertEqual(setting_str.encode("utf8"), dump_setting(res))

    def test_garage_settings(self):
        setting_str = '{"mode": "freeflow", "honest_payment_enabled": false}'
        res = parse_setting("common/garage_settings", setting_str)
        self.assert_result('{"mode": "freeflow", "honest_payment_enabled": true, "enforcement_enabled": true, '
                           '"payment_possible": true}', dump_setting(res))
        setting_str = '{"mode": "freeflow_surveillance"}'
        res = parse_setting("common/garage_settings", setting_str)
        self.assert_result('{"mode": "freeflow_surveillance", "honest_payment_enabled": false, '
                           '"enforcement_enabled": false, "payment_possible": false}', dump_setting(res))
        setting_str = '{"mode": "barrier"}'
        res = parse_setting("common/garage_settings", setting_str)
        self.assert_result('{"mode": "barrier", "honest_payment_enabled": false, "enforcement_enabled": false, '
                           '"payment_possible": true}', dump_setting(res))
        setting_str = '{"mode": "barrier_bypass"}'
        res = parse_setting("common/garage_settings", setting_str)
        self.assert_result('{"mode": "barrier_bypass", "honest_payment_enabled": false, "enforcement_enabled": false, '
                           '"payment_possible": false}', dump_setting(res))

    def test_location(self):
        setting_str = '{"name": "Garage", "street": "Garagestreet", "number": "12c", "floor": "2H", "door": "7a", ' \
                      '"supplements": "Frag nach Ines", "city": "Graz", "zip_code": "8020", "state": "Styria", ' \
                      '"country": "Österreich", "longitude": 15.441016366426302, "latitude": 47.041677687836156}'
        res = parse_setting("common/location", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_billing(self):
        setting_str = '{"billing_address": {"name": "Garage", "street": "Garagestreet", "number": "12c", ' \
                      '"floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz", ' \
                      '"zip_code": "8020", "state": "Styria", "country": "Österreich"}, "vat_id": "ATU1234567890"}'
        res = parse_setting("common/billing", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_support(self):
        setting_str = '{"name": "ARIVO", "address": {"name": "Garage", "street": "Garagestreet", "number": "12c", ' \
                      '"floor": "2H", "door": "7a", "supplements": "Frag nach Ines", "city": "Graz", ' \
                      '"zip_code": "8020", "state": "Styria", "country": "Österreich"}, ' \
                      '"phone_number": "+43 123 456", "email_address": "ines@arivo.co"}'
        res = parse_setting("common/support", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_garage_name(self):
        setting_str = '{"name": "Testgarage", "technical_name": "at-8-test-garage", "slug": "testgarage"}'
        res = parse_setting("common/garage_name", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_urls(self):
        setting_str = '{"payapp_domain": "pay.arivo.fun", "payapp_short_url": "https://pay.arivo.fun/arivo-v2", ' \
                      '"receipt_domain": "beleg.arivo.fun"}'
        res = parse_setting("common/urls", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_gate_control_mode(self):
        setting_str = '"standard"'
        res = parse_setting("gate_control/mode", setting_str)
        self.assertEqual(setting_str.encode("utf8"), dump_setting(res))

    def test_gate_control_day_mode(self):
        setting_str = '{"enabled": true, "start": "10:00", "end": "23:59"}'
        res = parse_setting("gate_control/day_mode", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_enforcement_basic_settings(self):
        setting_str = '{"strictness": 2, "payment_deadline_hours": 48, "enabled": true, ' \
                      '"ai_enabled": true, "last_edited": "2023-03-08T08:48:37.184000"}'
        res = parse_setting("enforcement/basic_settings", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_intercom_basic_settings(self):
        setting_str = '{"enabled": true, "phone_number": "sip:12345"}'
        res = parse_setting("intercom/basic_settings", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_device_keys_pairing(self):
        setting_str = '{"key": "arivopairing:ayek97rblos", "created_at": "2023-03-08T08:48:37.184000", "revision": 0}'
        res = parse_setting("device_keys/pairing", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_feature_flags(self):
        setting_str = '{"flag1": true, "oebb": true}'
        res = parse_setting("feature_flags", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_refs(self):
        setting_str = '{"name": "Parking Management System", "keys": ["2e9932bf-9071-47ac-afc5-3855a2eb6147", ' \
                      '"1f04cbeb-7613-496c-a301-6578eb2dafcd"]}'
        res = parse_setting("common/parksettings/refs/digimon", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_io_signals(self):
        setting_str = \
            """
            {"entry-open": {"signal": "entry-open", "gate": "entry", "type": "open", "name": "Einfahrt offen"},
             "exit-present_decision": {"signal": "exit-present_decision", "gate": "exit", "type": "present_decision", 
                                       "name": "Ausfahrt Entscheidung"}, 
             "00000000-0000-0000-0000-000000000001-area_full": 
                 {"signal": "00000000-0000-0000-0000-000000000001-area_full", "type": "area_full", 
                  "parking_area_id": "00000000-0000-0000-0000-000000000001", "name": "Parkbereich 1 voll"}, 
             "parkinglot_full": {"signal": "parkinglot_full", "type": "parkinglot_full", "name": "Parkplatz voll"}, 
             "custom_signal": {"signal": "custom_signal", "type": "custom", "gate": "entry", 
                               "parking_area_id": "00000000-0000-0000-0000-000000000001", "name": "Custom Signal"}, 
             "custom-signal2": {"signal": "custom-signal2", "type": "custom", "name": "Custom Signal 2"}}
            """
        res = parse_setting("io/signals", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_io_inputs(self):
        setting_str = \
            """
            {"entry-safety_loop": {"input":"entry-safety_loop","type":"safety_loop","gate":"entry","name":"EF CLoop",
            "active":true}, 
             "custom_input": {"input": "custom_input", "type": "custom", "gate": "entry", "name": "Custom Input",
             "active":true}, 
             "custom_input2": {"input": "custom_input2", "type": "custom", "name": "Custom Input 2","active":false}}
            """
        res = parse_setting("io/inputs", setting_str)
        self.assert_result(setting_str, dump_setting(res))

    def test_assets_visuals(self):
        setting_str = \
            """
            {"logo": {"type": "logo", "value": "stuff", "created": "2024-12-31T11:59:59"}, 
             "pogo": {"type": "pogo", "value": "hobo", "created": "2024-12-31T12:59:59"}}
            """
        res = parse_setting("assets/visuals", setting_str)
        self.assert_result(setting_str, dump_setting(res))


class EnsureAllTests(TestCase):
    def test_ensure_all_parsed(self):
        """Making sure all known settings are parsed at least once."""
        tests = [name for name, obj in ParsingTests.__dict__.items()
                 if isinstance(obj, FunctionType) and name.startswith("test_")]
        settings_keys = list(serialization._model_mapping.keys())
        self.assertGreaterEqual(len(tests), len(settings_keys))
        for key in settings_keys:
            if key.startswith("common/"):
                self.assertIn(f"test_{key.replace('common/', '')}", tests)
            else:
                self.assertIn(f"test_{key.replace('/', '_')}", tests)

    def test_ensure_all_docstrings(self):
        """Making sure all settings have a custom docstring."""

        def custom_docstring(settings_model: SettingsModel):
            return settings_model.__module__.startswith("settings_models") and \
                not any(base.__doc__ == settings_model.__doc__ for base in settings_model.__bases__)

        for key, setting in list(serialization._model_mapping.items()):
            if setting.settings_doc != setting.settings_model.__doc__:
                continue
            self.assertNotEqual(setting.settings_doc, "An enumeration.")
            if isinstance(setting.settings_model, SettingsModel):
                self.assertTrue(custom_docstring(setting.settings_model), msg=f"Missing docstring for {key}")
